#include "kernelsolver.h"

const int BLOCK_SIZE = 16;

__global__ void matrixMulCUDA(double *A, double *B, double *C, int wA, int wB) {
  int bx = blockIdx.x;
  int by = blockIdx.y;

  int tx = threadIdx.x;
  int ty = threadIdx.y;

  int aBegin = wA * BLOCK_SIZE * by;
  int aEnd   = aBegin + wA - 1;
  int aStep  = BLOCK_SIZE;

  int bBegin = BLOCK_SIZE * bx;
  int bStep  = BLOCK_SIZE * wB;

  double Csub = 0;
  for (int a = aBegin, b = bBegin; a <= aEnd; a += aStep, b += bStep) {
    __shared__ double As[BLOCK_SIZE][BLOCK_SIZE];
    __shared__ double Bs[BLOCK_SIZE][BLOCK_SIZE];

    As[ty][tx] = A[a + wA * ty + tx];
    Bs[ty][tx] = B[b + wB * ty + tx];

    __syncthreads();

    for (int k = 0; k < BLOCK_SIZE; ++k) {
      Csub += As[ty][k] * Bs[k][tx];
    }

    __syncthreads();
  }

  int c = wB * BLOCK_SIZE * by + BLOCK_SIZE * bx;
  C[c + wB * ty + tx] = Csub;
}

float solve(double *h_A, double *h_B, double *h_C, int size) {
  unsigned int matrixSize = size * size * sizeof(double);

  double *d_A, *d_B, *d_C;

  cudaMalloc((void **) &d_A, matrixSize);
  cudaMalloc((void **) &d_B, matrixSize);
  cudaMalloc((void **) &d_C, matrixSize);

  cudaMemcpy(d_A, h_A, matrixSize, cudaMemcpyHostToDevice);
  cudaMemcpy(d_B, h_B, matrixSize, cudaMemcpyHostToDevice);

  dim3 threads(BLOCK_SIZE, BLOCK_SIZE);
  dim3 grid(size / threads.x, size / threads.y);

  cudaEvent_t start, stop;
  float gpuTime = 0.0f;

  cudaEventCreate(&start);
  cudaEventCreate(&stop);
  cudaEventRecord(start, NULL);

  matrixMulCUDA<<< grid, threads >>>(d_A, d_B, d_C, size, size);

  cudaEventRecord(stop, NULL);
  cudaEventSynchronize(stop);
  cudaEventElapsedTime(&gpuTime, start, stop);

  cudaMemcpy(h_C, d_C, matrixSize, cudaMemcpyDeviceToHost);

  cudaFree(d_A);
  cudaFree(d_B);
  cudaFree(d_C);
  cudaDeviceReset();

  return gpuTime;
}
